\ FAST MULTI-TASKER using VERTICAL WINDOWING on 80C196KC

pto

Written by David Palmer of Nuclear Electric

This code is supplied by MPE without warranty.

This is a version of a multi-tasker which makes use of the vertical
windowing feature of the 80C196KC. It CANNOT be used with the previous
(-KA, -KB ) chips.

The vertical windowing provides many of the features of hardware bank
switching, allowing groups of physically distinct registers to be mapped
in turn into the same logical addresses. This enables context switching
to be achieved much faster and more elegantly than by multiple pops and
pushes. The INTEL data on the 80C196KC gives more information, and should be
consulted.

As detailed later the FORTH register equates (in MACROS.FTH) should
be changed. A new macros file is supplied as MACROSKC.FTH.

The word names used are the same as those used in MULTI96 and compiling
this file in place of the latter should allow previously developed programs
to run as before.

The main added words are :
KILLTASK and REVIVE-TASK which remove and restore tasks completely from
the scheduler.				     


\ The vertical windowing maps the 032H hex bytes from 0E0H to 0FFH
\ in the internal registers to register addresses:
\                                                 00 - 1F   V. window 0
\                    				  20 - 3F   "     "   1 
\						  40 - 5F   "     "   2
\ 						     |		 |
\						     |		 |
\						     |		 |
\						 1E0 - 1FF  V. window 0F

\ New Task Control Block ( TCB) layout
\ offsets into block
\ 0,1  ---- saved parameter stack pointer
\ 2    ---- task status byte  bit 7 (msb) 1 = active , 0 = halted
\ 3    ---- number of task sending message
\ 4,5  ---- message address
\ 6,7  ---- CFA of the main task word
\ 8,9  ---- CFA of the task event handler word

			\ multideb page 2 

hex

\ Equates for offsets into window follow:
 02 equ tskstat-offset
 03 equ msgorg-offset
 04 equ msgaddr-offset
 06 equ taskcfa-offset
 08 equ eventcfa-offset
 0A equ rp-offset	
 14 equ ip-offset
 16 equ up-offset 
 1E equ next-task       \ holds value to load into WSR to run next task
 1F equ self#-offset


\ Logical address equates as follows

 E0 equ tsksp		\ task stack pointer
 E2 equ tskstat         \ task status byte
 E3 equ msgorg          \ id of message sender
 E4 equ tskmes          \ message data
 E6 equ tskact          \ CFA of main task word
 E8 equ tskev           \ CFA of task event 

\ Status byte bit equates
 07 equ runbit			\ true = run task
 06 equ msgbit			\ true = message waiting
 05 equ evtrigbit               \ true = event triggered but not yet run
 04 equ evrunbit                \ true = event run or running

\ status byte bitmask equates
 080 equ runmask               \  bit 7 true = task running
 07F equ /runmask              
 040 equ msgmask               \  bit 6 true = message received
 03F equ /msgmask              
 020 equ evtrigmask            \  bit 5 true = event triggered
 0DF equ /evtrigmask
 010 equ evrunmask	       \  bit 4 true = event run
 0EF equ /evrunmask

\ * most register equates used by the FORTH system MUST BE CHANGED
\ * -- Alter " macros" file or use MACROSKC.FTH
\ * change initial values for tops of data and return stacks in
\ * MPB196KC.ctl to 1FE and 1C0 respectively
  
\ 018 constant SP	 --- data stack pointer, as before
\ 0EA constant RP        --- return stack pointer, was 01A
\ 0EC constant AX        --- was 01C 
\ 0EC constant AL      0ED constant AH
\ 0EE constant BX      0EE constant BL   0EF constant BH    --- from 01E....
\ 0F0 constant CX      0F0 constant CL   0F1 constant CH    --- from 020....
\ 0F2 constant DX      0F2 constant DL   0F3 constant DH    --- from 022....
\ 0F4 constant IP        --- Interpretive pointer, was 024
\ 0F6 constant UP        --- User area pointer, was 026
\ 0F8 constant TOS       --- top of data stack register, was 028
\ 0F8 constant tos.l  0F9 constant tos.h
\ 0FA constant tosext

			\ multideb page 3 /


\ the value to be loaded into the window select register (WSR) to context
\ switch to the next live task is stored at offset 01E in the window.
\ as stated above the 32 byte windows map logical addresses from 0E0-01F

  0FE constant tasklink		     \ WSR loading for next task
  0FF constant self#		     \ number of task in this window
 
\  Each task is run from its own set of registers, which are always
\ mapped in turn to the same set of logical addresses by writing the 
\ appropriate value to the WSR.
\ Task 0 is FORTH
\ Note that WSR values of 040 and 047 cannot be used since the former would 
\ map into the SFR's and the latter into the same physical locations
\ as task 0  ( which runs in a HORIZONTAL WINDOW )
\ WSR value 41 MAY be used as well if the system equates currently residing
\ in this block are moved.


 code task>window  ( tasknumber -- window )
      cmp tos r0
      ne, if,            	 \ task 0 runs in horizontal window 0
             add tos # 041       \ form window number, skipping 40, 41
             cmp tos # 046
             gt, if,
	             inc tos	 \ correct to skip window 47
	         endif,
          endif,
      next,
 end-code 

\ The WSR independent, address of each task's window base can be calculated
\ each time it is needed, but it is more efficient to have a table.
\ The number of entries can be reduced to the number required .

create tcb-table 00E0 , 0040 , 0060 , 0080 , 00A0 , 00C0 ,
                 0100 , 0120 , 0140 , 0160 , 0180 , 01AO ,

 code tcb-addr   ( task# -- n )    \ n = logical tcb base address of task#
	add tos tos		           \ word entries in table
        ld tos tcb-table tos ,[]	   \ get tcb address into tos
	next,
 end-code

 : tasklink-addr ( task# -- tasklink-addr of # )
                tcb-addr next-task +  	  \  addr. of register holding
 ;					  \ WSR value for NEXT task ( # + 1)
									    


\ allocate a table in RAM to hold the live tasks
\ the scheduler will only switch round live tasks
 
 THERE constant LIVE-TASKS  #tasks  allot-ram	    

\ The live task numbers are written to this table ( 0 = dead )
\ and the tasklink values adjusted accordingly.
\ The last task always links back to task 0.

			\ multideb page 4

 : clrlinks ( -- )                  \ clear all links
          #tasks 0 do 0 i 
                 tasklink-addr c!
              loop ;



 : set-links ( -- )	   \ reads live-tasks table and sets task links
     clrlinks                          \ wipe existing linkages
     0				       \ first task is 0
     #tasks 0 do live-tasks i + c@          \ get task no. from table
            dup 0= IF	               \ is it dead?
                     drop              \ yes ... then we needn't include it!
                   ELSE
		     dup	       \ we need to link to this task
                     task>window       \ WSR value to run task
                     rot               \ previous task
   		     tasklink-addr     \ get its tasklink address
                     c!		       \ link tasks
		   ENDIF
         loop
                0		    \ WSR value to run task 0
		swap
                tasklink-addr       \ tasklink addr. of final live task
                c!
;


 : livetask  ( # -- )               \ include task in live-task table but
           dup live-tasks + c!	    \ don't set links ( done by activate)
 ;  

\ KILLTASK and REVIVE-TASK  remove and replace tasks from the knowledge
\ of the scheduler. 
\ the scheduling of remaining tasks is faster but unlike a halted task
\ a dead task will not respond to events or messages until it is revived.
\ pending events or messages will then be dealt with.

 : killtask  ( # -- )               \ delete task from live-task table
           live-tasks +
           0 swap c!
	   set-links		    \ and remove from link table
 ;   

 : revive-task ( # -- )            \ restore task to live-task table
         livetask 
         set-links                 \ carry on as though never dead !
 ;


 : twipe   ( -- )         	    \ kill all tasks except 0
       #tasks 0 do i killtask loop
 ;


			\ multideb page 5 /

 : prlinks  ( -- )                   \ print tasklinks
	   cr
           #tasks 0 do i dup .		 \ print task#
                  tasklink-addr dup .    \ print its tasklink address
                  c@ .		         \ print window of task it links 
                  cr
              loop
 ;
	
\ PAUSE -- the scheduling word.
\ this checks to see whether multi-tasking is needed ( bit 7 in task# set )
\ if not it finishes.
\ otherwise it context switches to the next task in the linked task table
\ and checks for any active bits in the task status byte.
\ if none are set then it switches straight away to the next task
\ if any are set it loads the stack pointer for the task and then checks the
\ event triggered bit.
\ if this is set it first clears it and after setting the event running flag
\ runs the task event, followed by the continuing task action.
\ unless the event or another task clears the event running bit the next
\ time the scheduler services the task it will run the task action ( not the
\ event unless the event triggered bit has been set again) irrespective
\ of whether the task active bit ( bit 7) in the task status byte is clear.  
\ it will also, of course, run the task action if the task active or event
\ run bits are set.
 
 code pause	( -- )               \ main task switcher
	      jbs task# 7 [runtsk]   \ jump if single [9] else multi [5]
              st sp tsksp 	     \ save SP		[4]
l: [nxttask]  ldb wsr tasklink 	     \ switch to next task context [4]
	      cmpb tskstat r0        \ check status byte   [4]
              je [nxttask]	     \ zero? on to next task [8] else this [4]
	      ld sp tsksp	     \ load SP for this task  [4]
	      jbs tskstat evtrigbit [runev]    \  event if trig. [9], task [5] 
 l: [runtsk]     next,	           	       \		        [18]
 l: [runev]   andb tskstat # /evtrigmask       \ clear event trigger flag [5]
              orb  tskstat # evrunmask         \ mark event as running	  [5]
              br tskev []		       \ run event task		  [7]

 end-code


			\ multideb page 6 /


\ because context switching is now fast it may be worth removing the overhead
\ of checking for single/multi-tasking and always multi-task since the link
\ table will ensure that only desired tasks are polled.
\ However, the check for single/multitasking permits critical regions to
\ be implemented.
\ use of KILLTASK and REVIVE-TASK removes the need to check the task status
\ byte during the actual scheduling.
\ also if the application does not require tasks to run events then the
\ scheduling  word can be further simplified as in SCHEDULE below. 
\ events can then be looked for by the task itself, although the maximum
\ response latency could be high.


 code schedule
	       st sp tsksp                  \ must preserve SP  [4]
               ld wsr tasklink	            \ context switch  [4]  
               ld sp tsksp                  \ load this task's SP [4]
               next,                        \ run  it [18]
 end-code


			\ multideb page 7 /


\ assignment of words to perform the main and event tasks

 : to-task      ( taskcfa  task#  --  )
	tcb-addr                    \ tcb base address
        taskcfa-offset  	    \ offset into block to hold task cfa
        + !
 ;

 : to-event     ( eventcfa task# -- )
       tcb-addr			    \ tcb base address
       eventcfa-offset  	    \ offset into block to hold event cfa
       + !
  ;


\ activate makes a task known to the scheduler, including it in the
\ round-robin circuit. It initialises the TCB, marks the task as live 
\ and sets up the linkages.

: activate    ( task# --)  
	 dup
	 tcb-addr                     \ get window base address    
	 2dup 
        runmask over tskstat-offset +
         c!                           \ set run bit in status byte 
	 dup taskcfa-offset +	      \ addr. holding task cfa
         over ip-offset + !	      \ store in this task's IP
	 self#-offset +		      \ offset into window for storing task#
	 c!                           \ so store it!
         >r 			      \  save tcb base address on ret. stack
	 dup
	 per-task * init-u0 +         \ calculate address of task area in RAM
         task-init-s0 over +          \ task stack address
         2dup   		      \ 
         swap !			      \ set correct value of s0 in user area
	 2+                           \ add 2 to task stack address
         R@ 			      \   
         !                            \ store it in window task stack pointer
	 task-init-r0 over +          \ initial value for return pointer
         2dup
         swap 2+ !		      \ set return pointer in user area
         R@  rp-offset +  
         !                            \ store in window at rp-offset
	 r>  up-offset + 
         !                            \ store user area addr. in window	
         livetask                     \ mark task as live
         set-links                    \ and link with other tasks
 ;

 code single ( -- )    \  disable multi-tasker by setting bit 7 in task#
      orb task# # 080
      next,
 end-code

 code multi ( -- )     \  enable multi-tasker by clearing bit 7 in task#
     andb task# # 07F
     next,
 end-code

			\ multideb page 8 /


 code self    ( -- n )		 \  returns number of current task
          push tos
          ldbze tos.l self#
          next,
 end-code

 code status  ( -- tskstat  )  \ returns task status with runbit masked out
         push tos
	 ldbze tos.l tskstat
         andb tos.l # /runmask
	 next,
 end-code
  
 code halt ( task# -- )   \ clear run bit in status byte
	add tos tos		           \ word entries in table
        ld tos tcb-table tos ,[]	   \ get tcb address into tos
        ldbze al tskstat-offset tos ,[]	   \ status byte
        andb al # /runmask		   \ mask off run bit
        stb al tskstat-offset tos ,[]      \ store back
        pop tos				   \ clean up
        next,
 end-code      

 code sleep  ( task# -- )    \ clear status byte
	add tos tos		           \ word entries in table
        ld tos tcb-table tos ,[]	   \ get tcb address into tos
        stb r0   tskstat-offset tos ,[]	   \ status byte
	pop tos
        next,
 end-code

 : stop	     \  halt self
       self halt
       pause
 ;

 code restart ( task# -- ) \ set run bit in status byte
       add tos tos		           \ word entries in table
       ld tos tcb-table tos ,[]	           \ get tcb address into tos
       ldbze al tskstat-offset tos ,[]	   \ status byte
       orb al # runmask			   \ mask off run bit
       stb al tskstat-offset tos ,[]       \ store back
       pop tos				   \ clean up
       next,
 end-code      

 code set-event   ( task# --  )   \ set task#'s event flag
       add tos tos		           \ word entries in table
       ld tos tcb-table tos ,[]	           \ get tcb address into tos
       ldbze al tskstat-offset tos ,[]	   \ status byte
       orb al # evtrigmask		   \ set event trigger bit
       stb al tskstat-offset tos ,[]       \ store back
       pop tos				   \ clean up
       next,
 end-code      

			\ multideb page 9 /

 code clr-event-run  ( -- )        \ clear own event run flag
       andb tskstat # /evrunmask
       next,
 end-code

 code ?event      ( -- )          \ run own event task if triggered
     jbc tskstat evtrigbit [eventx]   \ if not triggered then carry on
     andb tskstat # /evtrigmask       \ clear event trigger bit
     orb  tskstat # evrunmask         \ set event run bit
     br tskev []                      \ run task event
 l: [eventx] 
	    next,
 end-code

code msg?  ( task# -- flag )    \ true if task has a message
	add tos tos		           \ word entries in tcb-table
        ld tos tcb-table tos ,[]	   \ get tcb address into tos
        ldbze al tskstat-offset tos ,[]	   \ status byte -> AL
	clr tos                            \ set flag false
        jbc al msgbit [msg?x]              \ end if no message
        dec tos				   \ set flag true
l: [msg?x]
        next,						  
 end-code
    
 code event?  ( task# -- flag )    \ true if task had an event trigger
	add tos tos		           \ word entries in tcb-table
        ld tos tcb-table tos ,[]	   \ get tcb base address into tos
        ldbze al tskstat-offset tos ,[]	   \ status byte -> AL
	clr tos                            \ set flag false
        jbc al evtrigbit [event?x]         \ end if no event
        dec tos				   \ set flag true
l: [event?x]
        next,
 end-code

 code [send-message]   ( message-addr task# -- ) \ store message in task
	add tos tos		            \ word entries in tcb-table
        ld tos tcb-table tos ,[]	    \ get tcb base address into tos
      	pop   msgaddr-offset tos ,[]        \ store message
	stb self# msgorg-offset tos ,[]     \ sender I.D
	ldb al   tskstat-offset tos ,[]     \ get status byte
        orb al # msgmask		    \ set message bit in status byte
        stb al tskstat-offset tos ,[]       \ store status byte 
	pop tos                             \ clean up
        next,
 end-code 

 code [get-message]   ( -- message-addr task# )      \ get the message
        push tos 
        push tskmes			    \ push message onto stack
        ldbze tos msgorg                    \ get I.D. of sender
        andb tskstat # 0BF                  \ clear message bit
        next,                               \ all done
 end-code


			\ multideb page 10 /


 : send-message   ( message-addr task# --  )   \ send message to task
  begin
       over msg?                       \ already got a message?
       while pause                     \ yes -- service other tasks
  repeat                               \ and then keep trying to pass message
 [send-message]                        \ pass on the message
 ; 

 : get-message  ( -- message-addr task#  )   \ wait for a message
  begin
       self msg? 0=                    \ does anybody love me ?
       while pause                     \ no -- then give other tasks a chance
  repeat			       \ and look again
  [get-message]			       \ news at last, read it
 ;


 : wait-event/msg               \ wait for message or event trigger
        begin pause status until ;

: wait	     ( n -- )          \ wait for n turns of scheduler or an event
      ?dup
	 if 0 do
              pause	       \ wait n times
              loop			     
        else wait-event/msg
        endif
 ;


			\ multideb page 11 /

\ semaphore
\
\ A SEMAPHORE is an extended variable used for signalling between tasks,
\ and for resource allocation.
\ The counter field is used as a count of the number of times the
\ resource may be used, and the arbiter field contains the TCB of the
\ task that last gained access. This field can be used for priority
\ arbitration and deadlock detection/arbitration.
\
: semaphore	\ -- ; -- addr [child]
  create  there 2+ ,			\ point to RAM
          0 ,(r)  0 ,(r)		\ count and arbiter fields
  ;code 
        pop ax push tos
        ld tos ax []			\ count field address
        next,
 end-code
   
\
\ SIGNAL increments the counter field of a semaphore, indicating
\ either a) that another item has been allocated to the resource, or
\        b) it is available for use again, 0 indicating in use by a task
\
 code signal	\ addr -- ; increment counter field of semaphore,
		\ so making it available for use by a task
  di 
      ld ax tos []	       \ get count field
      inc ax	               \ inc count field
      st ax tos []	       \ store back in semaphore
      pop tos		       \ clean up
  ei  
      next,
 end-code    


\
\ RELEASE is for use in REQUEST/RELEASE pairs for critical regions.
\ Because the cross compiler initialises data to zero, all semaphores
\ should be RELEASEd before use in the initialisation code.
\

 code release	\ addr -- ; set count field to one, and zero the owner field
 di
   ld ax # 01
   st ax tos []+	    \ set count field = 1
   st r0 tos []		    \ zero owner field
   pop tos		    \ clean up
 ei
  next,
 end-code

			\ multideb page 12 /

\ REQUEST waits until the counter field of a semaphore is non-zero,
\ and then decrements the counter field by one.
\ This allows the semaphore to be used as a COUNTED semaphore. For
\ example a character buffer may be used where the semaphore counter
\ shows the number of available characters.
\ Alternatively the semaphore may be used purely to share resources.
\ The semaphore is initialised to one. The first task to REQUEST it
\ gains access, and all other tasks must wait until the accessing task
\ SIGNALs that it has finished with the resource, or RELEASEs it.

 ' pause constant ps		     \ cfa of scheduler

 code request	( addr -- )      \ access a resource; wait if count = 0   
    di
      cmp r0 tos []		 \ test counter field of semaphore
      jne [askx]		 \ if zero then resource in use so wait
      ld ax # ps		 \ cfa of PAUSE
      dec ip dec ip		 \ reset I.P. so it will repeat this word
    ei				 \ when it is next scheduled
      br ax []			 \ execute scheduler
 l: [askx]
    ld ax tos []		 \ counter field not zero so got resource !
    dec ax			 \ decrement counter field
    st ax tos []+
    ldbze ax self#		 \ my I.D  ( task#, but could use wsr value ) 
    st ax  tos []		 \ and store it in arbiter field
    pop tos			 \ clean up
       ei
       next,
  end-code

\ a little test
\ prints message N times where N is value in counter field of flag
\ the flag can be loaded by other tasks

  align-odd
  semaphore flag 

 : test-request	( -- )     
  begin
        flag request ." got it ! " cr
  again 
 ;

			\ multideb page 13 /



: dummy-event	     \  print message if event requested
           ." event triggered in task " self . cr ;



: init-multi    \ -- ; initialisation with multi-tasking
            single			    \ disable multitasking

                                            \ * Set task user areas
       		                            \ * to a defined state

  [ init-u0 per-task + ] literal	    \ skip task 0
  [ #tasks per-task * ] literal  bounds	    \ but include interrupt task area
  do  i per-task erase			    \ clear space
      init-u0 i 040 cmove                   \ reset user areas
  per-task +loop			    \ in each task
  assign noop-task
 #tasks 1 do dup i tcb-addr 20 erase  	    \ clr tcb's except task 0
              i to-task 
       loop drop                            \ install safety task
   assign dummy-event
   #tasks 0 do dup i to-event
        loop drop                           \ install default event   
   assign abort 0 to-task
   0 tcb-addr 80 over tskstat-offset + c!   \  mark task 0 as running
   0 swap self#-offset + c!		    \  tell it who it is
   #tasks 1 do i dup activate halt loop	    \ set up all TCB's but don't run
 ; 


\ start off by including pause in i/o routine but the overhead is low enough
\ to sprinkle it around with little penalty.

: mt-key
  begin  (key?) 0=
    while  pause
  repeat
  (key) 
 ;
assign mt-key to-do key



			\ multideb ends /
\ the end

;p
