\ robot example for XC+ course

\ MicroProcessor Engineering Ltd
\ 133 Hill Lane
\ Southampton
\ SO1 5AF

\ Phone: 0703 631441

\ Version 1 - Written 1990 RGG to support robot driven by transistor switches
\ Version 2 - Rewritten 1992 GIE to support robot driven by multifunction
\	      I/O Board which in turn drives some relays


\ -----End of Page-----
\ vocabulary for robot words

vocabulary robot-defs


\ gie -- constants

hex

000 constant board			\ I/O board number
\ the above constant echoes the board select jumpers on the MFIO board

040 constant queue-size			\ how many instructions can we store
3 constant default-step			\ default step size: nominal tov counts per robot step

only forth also robot-defs definitions

0 constant cont	 			\ constant denoting continuous movement
				  
only forth definitions

\ word size independence page

\ target versions
: cell		\ -- #bytes ;
  2 ;

: cells		\ n -- #bytes ;
  2* ;

: cell+		\ addr -- addr' ;
  2+ ;

: cells+	\ addr n -- addr' ;
  2* + ;

\ host versions

host-compilation
only forth also c-c also interpreter cc/i

: cell cc/i		\ -- #bytes ;
  2 ;

: cells	cc/i		\ n -- #bytes ;
  2* ;

: cell+	cc/i		\ addr -- addr' ;
  2+ ;

: cells+ cc/i		\ addr n -- addr' ;
  2* + ;

only forth definitions
target-compilation

\ -----End of Page-----
\ gie --  variables

variable step-size			\ how big is a robot step?: timer ticks times steps required

\ these contain the values from the queue for the current movement
variable mask				\ value to write at the moment to port to cause this move
variable robot-move			\ the number of steps in the move we are making now

variable mask-queue queue-size cells allot-ram 	\ queue of port masks
variable step-queue queue-size cells allot-ram 	\ queue of step sizes

variable queue-in-pos			\ next free place in queue
variable queue-out-pos			\ next instruction to execute
variable queue-count 			\ how many in queue
 
\ -----End of Page-----
\ gie -- queue-handling extraction

: mask-queue>		\ -- mask ; get next mask to execute
  mask-queue queue-out-pos @ + @		\ fetch it
;

: step-queue>		\ -- step ; get next step to execute
  step-queue queue-out-pos @ + @		\ fetch it
;

: update-out-queue	\ -- step mask ; get instruction from queue
  step-queue> mask-queue>  		\ extract
  -1 queue-count +!			\ another item done
  queue-out-pos @ cell+ dup		\ increment position
  queue-size >				\ check limit (this is a circular queue)
  if
      drop				\ discard old position  
      queue-out-pos off			\ wrap
  else 
      queue-out-pos !			\ store new value
  endif
;

: queue>		\ -- step mask ; get instruction from queue							       
  queue-count @ 0=			\ check queue size limit
  if
    ." Error : No instruction in queue" cr 
    0 0					\ dummy instruction
  else
    update-out-queue			\ get instruction from queue 
  endif
;

\ gie -- queue-handling insertion

: >mask-queue		\ mask -- ; put mask in next free position
  mask-queue queue-in-pos @ + !		\ store it
;

: >step-queue		\ step -- ; put step in next free position
  step-queue queue-in-pos @ + !		\ store it
;

: update-in-queue	\ step mask -- ; insert instruction in queue
  >mask-queue >step-queue		\ insert
  1 queue-count +!			\ another item in queue
  queue-in-pos @ cell+ dup			\ increment position
  queue-size >				\ check limit (this is a circular queue)
  if
      drop				\ discard old position  
      queue-in-pos off			\ wrap
  else 
      queue-in-pos !			\ store new value
  endif
;

: >queue		\ mask step -- ; place instruction in queue							       
  queue-count @ cells queue-size <		\ check queue size limit
  if
    update-in-queue			\ place instruction in queue 
  else
    ." Instruction ignored : queue full " cr 
    2drop				\ ignore instruction
  endif
;

\ -----End of Page-----
\ -- R.G.G -- 10/05/90 --  direction -- HOST VERSION

also robot-defs definitions

host-compilation				\ host analogue of direction

only forth also c-c also interpreter cc/i	\ search order for def. word

required-def-word direction			\ need a new word type
init-target direction				\ start this word type

: direction   cc/i				\ n -- <name> ; -- n ;	
  ?0-EVEN  DUP  BL-WORD-UP CREATE		\ create entry in tar. dic.
  "" DR" HERE(T) .LOG				\ say so, as compiler runs
  target direction compile-run			\ what to lay down in target
  ,(T)						\ comma in value
  make-c/v-data-packet				\ this is a data packet
;

only forth definitions				\ back to this order	

target-compilation				\ and back to the target

only forth definitions

\ -----End of Page-----
\ -- R.G.G -- 10/05/90 --  direction -- TARGET VERSION

only forth also robot-defs definitions

: direction			\ n -- <name> ; -- n ;
  create
    ,
  does>
    @
;

\ -----End of Page-----
\ gie --  axis

: axis
  create      		\ cw/up aw/down -- ; build data structure
    ,					\ value required to move cw/up
    ,					\ value to move aw/down
  does>			\ steps dir -- ; set up and kick timer
    swap cells+ @ 			\ the mask to use
    >queue				\ put it in the queue
;
 
\ -----End of Page-----
\ gie -- axis control and directions
hex

\ active high
\ mask mask axis name			\ axis number on robot		 

00200 00100 axis torso 			\ the base (axis 1)
01000 02000 axis shoulder		\ the third axis
00800 00400 axis elbow	 		\ the second axis
08000 04000 axis wrist 			\ the fourth axis
00002 00001 axis hand			\ the fifth axis
00000 00000 axis rest			\ dummy stop

\ index direction name
0 direction open			\ open end-effector
1 direction close			\ close end-effector

0 direction up				\ up
1 direction down			\ down

0 direction cw				\ clockwise
1 direction aw				\ anticlockwise

0 direction null			\ just for <nn> NULL REST

only forth definitions

\ -----End of Page-----
\ gie -- timer start and stop
							  
\ Note that some of the interrupt driver code is in INT196.FTH

: start-timer	\ -- ; zero counters and enable interrupts
  tov-ticks off				\ zero the counter
  start-tov				\ kick off the interrupts
;

: stop-timer     \ -- ; disable overflow interrupts
  0 int_mask c!				\ disable TOV interrupts
;

\ gie -- stop-robot kill

hex		    

: stop-robot		 \ -- ; stop current robot movement
  0 board o!				\ turn off robot axes
  stop-timer				\ stop interrupts
  robot-move off			\ zero the robot movement counter
  tov-ticks off				\ zero the ticks counter
  mask off 				\ clear the mask
;

\ only forth also robot-defs definitions

: kill 			 \ -- ; stop current movement and kill task
  stop-robot	 			\ turn off movement 
  1 halt				\ and turn off robot task
  queue-count off			\ pretend no more instructions in queue
  queue-in-pos off			\ ditto
  queue-out-pos off
;		    

\ only forth definitions

\ -----End of Page-----
\ gie --  robot-movement low-level

: cont-move		\ -- finished? ; move robot until key press
  mask @ board o!			\ set mask
  begin  key?  until  			\ wait for key press
  stop-robot 	  			\ stop it
  -1					\ finished instruction
;

: normal-move 		\ -- finished? ; move robot while necessary
  robot-move @ step-size @ *  
  tov-ticks @ > 	     		\ does the robot still need to move ?
  if
    mask @ board o!			\ continue movement
    0					\ not finished
  else
    stop-robot				\ stop movement
    -1					\ finished instruction
  endif
;

only forth also robot-defs also forth definitions	\ need "cont" in this word

: move-robot		\ -- ; move robot 1 step if required
  begin
    robot-move @		\ how far does the robot have to move?
    cont =			\ continuous?
    if  			\ YES:
      cont-move			\    move continuously 
    else		   	\ NO:
      normal-move		\    move till done		
    endif					      
    pause			\ reschedule
  until				\ exit when instruction finished so we can
				\ get the next one from the queue
;
						     
only forth definitions

((
NB The cont = ... structure is less efficient than the equivalent
 robot-move if, but the former is more descriptive of the requirement
))

\ -----End of Page-----
\ gie -- instruction-exec

: instruction-exec	\ -- ; move instruction from queue to execution 
  queue>				\ get instruction
  mask !				\ stash the port mask
  robot-move !				\ and the length of movement			
  start-timer				\ kick off timer
  move-robot				\ and make the movement
;

\ -----End of Page-----
\ gie -- check-queue

: check-queue		\ -- ; if anything in queue - run it
  queue-count @				\ count > 0 ?
  if  instruction-exec  endif		\ yes : run instruction
;

\ -----End of Page-----
\ gie -- robot-task

: robot-task  		\ -- ; action of delay task
  begin		     
    check-queue				\ run anything in the inst queue
    pause				\ and schedule for main task
  again	 				\ continuous loop
;
\ -- R.G.G -- 11/05/90 --  motion & end-motion

only forth also robot-defs definitions

: motion				\ -- ;
  [compile] :						\ same action as :
;

: end-motion			\ -- ;
  [compile] ;						\ same action as ;
;  immediate						\ and must be immediate

only forth definitions

\ -----End of Page-----
\ robot pretty words and forth access

decimal   

only forth also robot-defs definitions

: done			\ -- ; 
  kill					\ turn off robot
  only forth definitions		\ and back to interactive forth
  assign (prompt) to-do prompt		\ back to normal prompt
;

: help			\ -- ; so the system is complete
  cr ." The following commands are available:" cr
  words cr
;

only forth definitions

: r-prompt		\ -- ; alternative prompt
  ." robot: "
;

: signon		\ -- ; 
  cr ." MPE Robot Control Langauge Demonstrator."
  cr ." (c) MicroProcessor Engineering Ltd., 1992"
  cr cr ." Type `Help' for a list of available commands"
  cr cr
  space r-prompt
;
\ gie --  initialise controller

: init-robot		\ -- ;
  stop-robot				\ ensure robot halted
  queue-in-pos off			\ initialise variables
  queue-out-pos off
  queue-count off
  default-step step-size !		\ initalise step - magic multiplier
  assign r-prompt to-do prompt		\ and use new prompt
;

\ top level word
		   
: robot				\ -- ;
  signon
  init-robot
  init-multi	 			\ initialise system
  multi					\ start multi-tasker
  assign robot-task 1 to-task		\ assign action
  1 activate				\ start timer task
  only robot-defs definitions		\ use only selected words
  abort					\ go run the thing
;

assign robot to-do loader		\ **** to make turnkey robot application
